package controllers

import dao.{TestCaseDAO, TestRunDAO}
import models.TestCase
import play.api.data.Forms._
import play.api.data._
import play.api.data.format.Formatter
import play.api.mvc.{Action, Controller}
import services.TestRunner

import scala.xml.Node

/**
  * Main controller for the testsuite. Answers all the request.
  */
object ApplicationController extends Controller {

  val testDao = new TestCaseDAO()
  val testRunDAO = new TestRunDAO
  val testRunner = new TestRunner()
  /**
    * Gets calles when the base url of the server is requested.
    * Redirects to ApplicationController.index.
    */
  def root = Action {
    Redirect(routes.ApplicationController.index())
  }

  /**
    * The landing page of the test suite.
    */
  def index = Action {
    Ok(views.html.index(runTestForm, testRunDAO.all().nonEmpty))
  }

  /**
    * Initiates a test run for the base url of the cts system that was defined in the form.
    */
  def runTest = Action { implicit request =>
    runTestForm.bindFromRequest.fold(
      hasErrors => {
        BadRequest(views.html.index(hasErrors, testRunDAO.all().nonEmpty))
      },
      runTestData => {
        val testRun = testRunDAO.add(testRunner.run(testDao.all(), runTestData.baseUrl))
        Redirect(routes.ApplicationController.testRun(testRun.id.get))
      }
    )
  }

  /**
    * Helper object to get the value of the base url field in the form.
    */
  val runTestForm = Form(
    mapping("baseurl" -> nonEmptyText
    )(RunTestData.apply)(RunTestData.unapply)
  )


  /**
    * Returns a view with all available tests.
    */
  def tests = Action {
    Ok(views.html.tests(testDao.all()))
  }

  /**
    * Returns the detail page of the test referenced by its id.
    * @param id id of the TestCase
    */
  def test(id:String) = Action {
    testDao.findById(id) match {
      case Some(testCase) => Ok(views.html.test(testCase))
      case None => Redirect(routes.ApplicationController.tests())
    }
  }

  /**
    * Makes the server reload the defined tests from disk.
    */
  def reloadTests() = Action {
    testDao.reload()
    Ok(views.html.tests(testDao.all()))
  }

  /**
    * Helper object to extract the field from the form for editing/creating tests.
    */
  val testCaseForm = Form(
    mapping("id" -> nonEmptyText,
      "name" -> nonEmptyText,
      "description" -> nonEmptyText,
      "requesttype" -> nonEmptyText,
      "additionalparameters" -> stringSeq,
      "urn" -> nonEmptyText,
      "expectedresponse" -> xmlNode
    )(TestCase.apply)(TestCase.unapply)
  )

  /**
    * Shows the form to edit a test if a valid id was passed.
    * @param id id of the TestCase
    */
  def showEditTestForm(id:String) = Action {
    testDao.findById(id) match {
      case Some(testCase) => Ok(views.html.editTestForm(testCase))
      case None => Redirect(routes.ApplicationController.tests())
    }
  }

  /**
    * Updates the existing test with the values passed from the edit form.
    * @param id id of the TestCase
    */
  def updateTest(id:String) = Action { implicit request =>
    testCaseForm.bindFromRequest().fold(
      hasErrors => {
        BadRequest(views.html.editTestForm(testDao.findById(id).get))
      },
      testCaseData => {
        testDao.update(testCaseData)
        Redirect(routes.ApplicationController.test(testCaseData.id))
      }
    )
  }

  /**
    * Shows the form to create a new test.
    */
  def showNewTestForm() = Action {
    Ok(views.html.newTestForm())
  }

  /**
    * Extracts the values for the new test from the form an stores it on the disk.
    */
  def addTest() = Action { implicit request =>
    testCaseForm.bindFromRequest().fold(
      hasErrors => {
        BadRequest(views.html.newTestForm())
      },
      testCaseData => {
        testDao.add(testCaseData)
        Redirect(routes.ApplicationController.test(testCaseData.id))
      }
    )
  }

  /**
    * Shows a list with all test runs that have been made.
    */
  def testRuns = Action {
    Ok(views.html.testruns(testRunDAO.all()))
  }

  /**
    * Shows the detail page of a test run.
    * @param id id of the TestRun
    */
  def testRun(id:Long) = Action {
    testRunDAO.findById(id) match {
      case Some(testRun) => Ok(views.html.testrun(testRun))
      case None => Redirect(routes.ApplicationController.index())
    }
  }

  /**
    * A helper function to go directly to the detail page of the latest test run.
    */
  def latestTestRun() = Action {
    if (testRunDAO.all().isEmpty) {
      Redirect(routes.ApplicationController.index())
    } else {
      Redirect(routes.ApplicationController.testRun(testRunDAO.all().sortBy(_.id).last.id.get))
    }
  }

  /**
    * Shows the detail page of the result for a certain test result.
    * @param runId id of the TestRun
    * @param resultId id of the TestResult
    */
  def testResult(runId:Long, resultId:Long) = Action {
    testRunDAO.findById(runId) match {
      case Some(testRun) =>
        testRun.results.find(_.id == resultId) match {
          case Some(testResult) => Ok(views.html.testResult(testResult))
          case None => Redirect(routes.ApplicationController.testRuns())
        }
      case None => Redirect(routes.ApplicationController.testRuns())
    }
  }

  /**
    * Helper functions to interpret certain form fields as xml values.
    */
  def xmlNode = of(xmlNodeFormat)
  def xmlNodeFormat: Formatter[Node] = new Formatter[Node] {
    def bind(key: String, data: Map[String, String]) = {
      try {
        Right(scala.xml.XML.loadString(data.getOrElse(key, "")))
      } catch {
        case e: Exception => Left(List(FormError(key, "Could not parse string to xml.")))
      }
    }
    def unbind(key: String, value: Node) = Map(key -> "")
  }

  /**
    * Helper functions to interpret certain form fields as a sequence of pairs separated by ';'
    */
  def stringSeq = of(stringSeqFormat)
  def stringSeqFormat: Formatter[Seq[String]] = new Formatter[Seq[String]] {
    def bind(key: String, data: Map[String, String]) = {
      try {
        Right(data.getOrElse(key, "").split(";").toList)
      } catch {
        case e: Exception => Left(List(FormError(key, "Could not split string.")))
      }
    }
    def unbind(key: String, value: Seq[String]) = Map(key -> "")
  }
}

case class RunTestData(baseUrl:String)


